import pygame
from pygame.math import Vector2
from const import *
import numpy as np
from abc import abstractmethod
import math


class Agent(pygame.sprite.Sprite):
    def __init__(self, img: str, size: tuple, v_drct: Vector2, v_norm: float,
                 a_drct: Vector2, a_norm: float, m: float, init_pos: tuple, id=None,
                 name=None, potential_treasure=None):
        pygame.sprite.Sprite.__init__(self)
        player_img = pygame.image.load(img).convert()
        self.image = pygame.transform.scale(player_img, size)
        self.image.set_colorkey(WHITE)
        self.rect = self.image.get_rect()
        self.pos = Vector2(init_pos[0], init_pos[1])
        self.v_drct = v_drct  # 速度方向
        self.v_norm = v_norm  # 速度模长
        self.a_drct = a_drct  # 加速度方向
        self.a_norm = a_norm  # 加速度模长
        self.m = m  # 相对质量
        self.rect.center = init_pos
        self.radius = size[0] / 2
        self.size = size

        self.id = id
        self.name = name

        self.own_ids = []
        self.holding_ids = []

        self.intent = None

        self.speed = 4.5
        self.speed_arr = [4.5, 3]  # attacker所有速度的可能性，[0]为没有抱着物体时的速度，[1]为抱着时的速度
        self.potential_treasure = potential_treasure
        self.treasure = None
        self.get_collision: bool = False

        self.circle_speed = 4.5 # 做圆周运动的速度
        self.toward_speed = 4.5 # toward函数中的速度
        self.toward_a_norm = 0.1 # toward函数中调整速度所用的加速度

    def update(self):
        # 更新intent
        leaf_intent = self.intent.get_leaf_intent()
        leaf_intent.update()
        # print("update_intent")
        # intent-action
        leaf_intent = self.intent.get_leaf_intent()
        leaf_intent.update_action()
        # print(leaf_intent.name)
        self.ahead()
        # 检测碰撞
        self.collision_detection()
        return super().update()

    def init_intent(self, world):
        self.intent = Intent(self, world, "survive")
        self.world = world

    def ahead(self):
        a: Vector2 = self.a_drct * self.a_norm
        v: Vector2 = self.v_drct * self.v_norm + a
        if v.length() != 0:
            self.v_drct = v.normalize()
        self.v_norm = v.length()
        self.pos = self.pos + v
        self.rect.centerx = self.pos[0]
        self.rect.centery = self.pos[1]

    def collision_detection(self):
        if self.rect.bottom >= WINDOWS_H or self.rect.top <= 0:
            self.v_drct[1] = -self.v_drct[1]
        if self.rect.right >= WINDOWS_W or self.rect.left <= 0:
            self.v_drct[0] = -self.v_drct[0]

    def toward_a(self, dest: Vector2):
        path = dest - Vector2(self.rect.centerx, self.rect.centery)
        path_lenth = path.length()
        if path_lenth < self.v_norm:
            self.rect.center = dest
        elif path_lenth != 0:
            self.a_drct = path_lenth

    def bounce(self, agent):
        v: Vector2 = ((self.m - agent.m) * self.v_norm * self.v_drct + 2 * agent.m * agent.v_norm * agent.v_drct) / (
            self.m + agent.m)
        v_agent: Vector2 = ((
            agent.m - self.m) * agent.v_norm * agent.v_drct + 2 * self.m * self.v_norm * self.v_drct) / (
            self.m + agent.m)
        if v.length() != 0:
            self.v_drct = v.normalize()
        self.v_norm = v.length()
        agent.v_drct = v_agent.normalize()
        agent.v_norm = v_agent.length()
        if agent.treasure:
            agent.get_collision = True

    def own(self, id):
        return id in self.own_ids

    def hold_something(self):
        return bool(len(self.holding_ids))

    def get(self, id):
        if not self.hold_something():
            if id not in self.holding_ids:
                self.holding_ids.append(id)
                return True
        return False

    def put_down(self, id):
        if id in self.holding_ids:
            self.holding_ids.remove(id)
            return True
        return False

    def get_position(self):
        return self.rect.center

    def get_Treasure(self, treasure):
        GET_DIS = 25
        if self.treasure:
            if self.get_collision:
                if not pygame.sprite.collide_circle(self, self.treasure):
                    self.treasure = None
                    self.get_collision = False
            else:
                self.treasure.rect.center = self.rect.center
        else:
            if (not self.get_collision) and \
                    (self.rect.center[0] - treasure.rect.center[0]) ** 2 + (
                    self.rect.center[1] - treasure.rect.center[1]) ** 2 < GET_DIS:
                self.treasure = treasure.id
                treasure.being_held_id = self.id
        print("treasure:", self.treasure == None)
        print("self.getcollosion", self.get_collision)

    def has_treasure(self):
        if self.treasure is None:
            return False
        else:
            return True

    def toward_v(self, dest: Vector2):
        path = dest - self.pos
        path_lenth = path.length()
        if path_lenth < self.v_norm or path_lenth < 1.415:
            self.rect.center = dest
            self.pos = dest
            self.v_norm = 0
            self.a_norm = 0
        elif path_lenth > 1.415:
            self.v_drct = path.normalize()
            self.v_norm = self.speed

    # def circle(self, center: Vector2):
    #     """
    #     围绕指定点画圆(默认self已经处于圆周上了)
    #     这个函数和ahead不能够同时使用，这里不是ahead的模式！！！这个函数是直接通过位置的改变而运动的
    #     """
    #     center = Vector2(center)
    #     radius: Vector2 = self.pos - center
    #     rotated_radius = radius.rotate_rad(self.circle_speed / math.sqrt(radius[0]**2 + radius[1]**2))
    #     # radius.rotate_ip_rad(self.circle_speed / math.sqrt(radius[0]**2 + radius[1]**2)) # 以弧度为单位旋转，速度可调
    #     print(rotated_radius, radius)
    #     self.pos = center + rotated_radius

    def circle(self, center : Vector2):
        """
        改版后的circle函数
        """
        center = Vector2(center)
        reverse_radius: Vector2 = center - self.pos 
        radius_len = reverse_radius.length()
        reverse_radius.normalize_ip()
        self.a_drct = reverse_radius
        self.a_norm = self.circle_speed ** 2 / radius_len
        self.v_norm = self.circle_speed
        self.v_drct = reverse_radius.rotate(90)
        

    def defend(self, center: Vector2, radius: float, enemy: Vector2):
        """
        防守函数，前往距离center和enemy之间的，与center具有指定距离radius的点
        如果center和enemy间的距离小于radius，则前往center和enemy的中点
        """
        center = Vector2(center)
        enemy = Vector2(enemy)
        distance : Vector2 = enemy - center
        if distance.length() < radius:
            self.toward(center + distance / 2)
        else:
            self.toward(center + distance.normalize() * radius)

    def toward(self, dest: Vector2):
        """
        朝指定点前进
        """
        # print(dest)
        dest = Vector2(dest)
        path = dest - self.pos
        v_toward = dest - self.pos
        v_toward.scale_to_length(self.toward_speed) # 追赶时所需要的速度
        v = self.v_norm * self.v_drct
        if v_toward != v: # 如果没到追赶速度的时候，进行速度改变
            a = v_toward - v
            if a.length() < self.toward_a_norm:
                self.a_drct = a.normalize() # a = v_toward - v，这样一个运动周期后v_toward = v
                self.a_norm = a.length()
            else:
                a.scale_to_length(self.toward_a_norm)
                self.a_drct = a.normalize()
                self.a_norm = a.length()
        else:
            path_lenth = path.length()
            if path_lenth < self.v_norm or path_lenth < 1.415:
                self.rect.center = dest
                self.pos = dest
                self.v_norm = 0
                self.a_norm = 0
            elif path_lenth > 1.415:
                self.v_drct = path.normalize()
                self.v_norm = self.speed
                self.a_norm = 0

    def be_bounced(self):
        if self.treasure is None:
            return
        treasure = self.world.retrieve_by_id(self.treasure)
        treasure.being_held_id = None
        self.treasure = None


from intent import Intent

