from utlis import DEFEND_RADIUS, get_norm_direction, distance, SAFE_DISTANCE, is_near


# 这是为了构建基本框架及制作基本demo的简化版intent
# 由于我们的agent是全知的，intent的变化只取决于对对方mental states的猜测
# 每个单独的Intent是intent-structure上的一个节点，它们共同组成了Intent树
# 这里的intent作简化处理，沿着固定的规则进行转化，且每个intent有着几乎确定的action sequence
# 已知的所有intent及其转变规则
# defend treasure -> wander around treasure(for detecting)(if detect no attacker)
#                 -> hinder attacker(if the agent's intent is attacking(guess))
# get treasure    -> head for treasure
# 这里简化了task和intent的区别，task被视为最底层的intent，其对应着多种policy
class Intent:
    def __init__(self, agent, world, name=None):
        self.name = name
        self.agent = agent
        self.world = world
        self.parent_intent = None
        self.child_intent = None
        self.pre_child_intent = []

    # 更新intent
    # intent-graph 暂时用if-else模拟，每一个if相当于graph中的一条边
    # status: 表示各种状态的类
    # 每次转移成功后需要递归地调用update函数，以保证一直迭代到最后一层可执行的intent，或者抛出异常
    def update(self):
        if self.name == "survive":
            if self.world.has_treasure(self.agent) and not self.world.lose_treasure(self.agent):
                sub_intent = Intent(self.agent, self.world, "defend treasure")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
            else:
                sub_intent = Intent(self.agent, self.world, "get treasure")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
        elif self.name == "defend treasure":
            if self.world.lose_treasure(self.agent):
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
            elif self.world.detect_attacker(self.agent):
                sub_intent = Intent(self.agent, self.world, "hinder attacker")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
            else:
                sub_intent = Intent(self.agent, self.world, "wander around treasure")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
        elif self.name == "get treasure":
            if self.world.has_treasure(self.agent) and not self.world.lose_treasure(self.agent):
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
            if self.agent.has_treasure():
                sub_intent = Intent(self.agent, self.world, "get rid of others")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
            elif self.world.detect_treasure_on_agent(self.agent):
                sub_intent = Intent(self.agent, self.world, "attack")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
            else:
                sub_intent = Intent(self.agent, self.world, "move to treasure")
                self.add_child_intent(sub_intent)
                sub_intent.update()
                return True
        elif self.name == "hinder attacker":
            if not self.world.detect_attacker(self.agent):
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
        elif self.name == "wander around treasure":
            if self.world.detect_attacker(self.agent):
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
        elif self.name == "attack":
            if self.world.detect_treasure_on_ground(self.agent):
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
        elif self.name == "move to treasure":
            if self.agent.has_treasure():
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
        elif self.name == "get rid of others":
            if self.world.has_treasure(self.agent):
                # 回溯
                parent_intent = self.parent_intent
                parent_intent.delete_child_intent(self)
                parent_intent.update()
                return True
        return False

    # intent 向下一层转移
    def add_child_intent(self, sub_intent):
        sub_intent.parent_intent = self
        self.child_intent = sub_intent

    # intent 进行回溯
    def delete_child_intent(self, sub_intent):
        self.pre_child_intent.append(self.child_intent)
        self.child_intent = None

    # 从叶节点更新动作
    def update_action(self):
        if self.name == "wander around treasure":
            treasure_id = self.world.owner2object[self.agent.id]
            treasure = self.world.retrieve_by_id(treasure_id)
            distance_ = distance(treasure, self.agent)
            treasure_position = treasure.get_position()
            if -20 < abs(distance_ - DEFEND_RADIUS) < 20:
                return self.agent.circle(treasure_position)
            else:
                norm_direction = get_norm_direction(self.agent, treasure)
                # print("treasure_position", treasure.get_position())
                # print("agent_position", self.agent.get_position())
                # print("norm_direction", norm_direction)
                target_position = (treasure_position[0] + abs(norm_direction[0]) * DEFEND_RADIUS,
                                   treasure_position[1] + abs(norm_direction[1]) * DEFEND_RADIUS)
                return self.agent.toward(target_position)
        elif self.name == "hinder attacker":
            attacker = self.world.detect_attacker(self.agent)
            treasure_id = self.world.owner2object[self.agent.id]
            treasure = self.world.retrieve_by_id(treasure_id)
            treasure_position = treasure.get_position()
            return self.agent.defend(treasure_position, DEFEND_RADIUS, attacker.get_position())
        elif self.name == "attack":
            attacker_id = self.world.detect_treasure_on_agent(self.agent)
            attacker = self.world.retrieve_by_id(attacker_id)
            return self.agent.toward(attacker.get_position())
        elif self.name == "move to treasure":
            treasure = self.world.nearest_treasure(self.agent)
            if is_near(treasure, self.agent):
                return self.agent.get_Treasure(treasure)
            return self.agent.toward(treasure.get_position())
        elif self.name == "get rid of others":
            near_agent = self.world.nearest_agent(self.agent)
            norm_direction = get_norm_direction(self.agent, near_agent)
            target_position = (near_agent.get_position()[0] + norm_direction[0] * SAFE_DISTANCE,
                               near_agent.get_position()[1] + norm_direction[1] * SAFE_DISTANCE)
            return self.agent.toward(target_position)

    def get_leaf_intent(self):
        temp = self
        while temp.child_intent is not None:
            temp = temp.child_intent
        return temp
