# 保存所有的agents, objects 和 landmarks
from utlis import distance, is_near
import numpy as np
import math


class World():
    def __init__(self):
        self.agents = []
        self.objects = []
        self.landmarks = []

        self.object2owner = {}
        self.owner2object = {}

    def add_agent(self, target):
        self.agents.append(target)

    def add_object(self, target):
        self.objects.append(target)

    def add_landmark(self, target):
        self.landmarks.append(target)

    def can_put_down(self, target):
        for object in self.objects:
            if is_near(object, target):
                return False
        return True

    # 对于目标agent来说，有没有检测到威胁
    def detect_attacker(self, agent):
        for temp in self.agents:
            if temp.id == agent.id:
                continue
            direction = np.asarray(agent.get_position()) - np.asarray(temp.get_position())
            norm_direction = direction / math.sqrt(direction[0]**2 + direction[1]**2)
            v = temp.v_drct
            cos = v[0] * norm_direction[0] + v[1] * norm_direction[1]
            if -0.5 < cos < 0.5:
                return temp
        return False

    # 初始化ownership关系
    # object_id: owner.id
    def init_treasure(self):
        owner_ids = []
        for i in range(len(self.objects)):
            all_distance = []
            for j in range(len(self.agents)):
                all_distance.append(distance(self.agents[j], self.objects[i]))
            all_distance = np.asarray(all_distance)
            index = np.argsort(all_distance)
            k = 0
            while self.agents[index[k]].id in owner_ids:
                k += 1
            self.object2owner[self.objects[i].id] = self.agents[index[k]].id
            self.owner2object[self.agents[index[k]].id] = self.objects[i].id
            owner_ids.append(self.agents[index[k]].id)

    # None: 不拥有
    # id: 拥有的id
    def has_treasure(self, agent):
        if self.owner2object.get(agent.id) is not None:
            return True
        else:
            return False

    def lose_treasure(self, agent):
        if not self.has_treasure(agent):
            return True
        treasure_id = self.owner2object[agent.id]
        treasure = self.retrieve_by_id(treasure_id)
        if treasure.being_held_id is not None and treasure.being_held_id != agent.id:
            return True
        return False

    def retrieve_by_id(self, id):
        for agent in self.agents:
            if agent.id == id:
                return agent
        for object in self.objects:
            if object.id == id:
                return object

    # 假设agent不含有treasure
    def detect_treasure_on_agent(self, agent):
        # 先找出离agent最近的object
        all_distance = []
        for object in self.objects:
            all_distance.append(distance(agent, object))
        all_distance = np.asarray(all_distance)
        index = np.argsort(all_distance)
        treasure = self.objects[index[0]]
        if treasure.being_held_id is not None:
            return treasure.being_held_id
        return False

    def detect_treasure_on_ground(self, agent):
        return not self.detect_treasure_on_agent(agent)

    def nearest_treasure(self, agent):
        # 先找出离agent最近的object
        all_distance = []
        for object in self.objects:
            all_distance.append(distance(agent, object))
        all_distance = np.asarray(all_distance)
        index = np.argsort(all_distance)
        treasure = self.objects[index[0]]
        return treasure

    def nearest_agent(self, agent):
        all_distance = []
        for temp in self.agents:
            if temp.id == agent.id:
                all_distance.append(50000)
                continue
            all_distance.append(distance(temp, agent))
        all_distance = np.asarray(all_distance)
        index = np.argsort(all_distance)
        return self.agents[index[0]]
